@@ -19,18 +19,25 @@ import tempfile
 from oslo_concurrency import processutils as putils
 from oslo_log import log as logging
 from oslo_serialization import jsonutils
+from oslo_utils import excutils
 from oslo_utils import fileutils
 from oslo_utils import netutils
+from oslo_utils import versionutils
 
 from os_brick import exception
 from os_brick.i18n import _
 from os_brick import initiator
 from os_brick.initiator.connectors import base
 from os_brick.initiator import linuxrbd
+from os_brick.privileged import rbd as rbd_privsep
 from os_brick import utils
 
 LOG = logging.getLogger(__name__)
 
+# BUG: #1865754 - Set version base for ceph Octopus which requires additional
+# handling of the config file.
+OCTOPUS_BASE_VERSION = "15.2.0"
+
 
 class RBDConnector(base.BaseLinuxConnector):
     """"Connector class to attach/detach RBD volumes."""
@@ -63,14 +70,16 @@ class RBDConnector(base.BaseLinuxConnector):
         # TODO(e0ne): Implement this for local volume.
         return []
 
-    def _sanitize_mon_hosts(self, hosts):
+    @staticmethod
+    def _sanitize_mon_hosts(hosts):
         def _sanitize_host(host):
             if netutils.is_valid_ipv6(host):
                 host = '[%s]' % host
             return host
         return list(map(_sanitize_host, hosts))
 
-    def _check_or_get_keyring_contents(self, keyring, cluster_name, user):
+    @staticmethod
+    def _check_or_get_keyring_contents(keyring, cluster_name, user):
         try:
             if keyring is None:
                 if user:
@@ -94,15 +103,42 @@ class RBDConnector(base.BaseLinuxConnector):
         keyring = self._check_or_get_keyring_contents(keyring, cluster_name,
                                                       user)
 
+        ceph_version = self._determine_ceph_client_version()
         try:
             fd, ceph_conf_path = tempfile.mkstemp(prefix="brickrbd_")
             with os.fdopen(fd, 'w') as conf_file:
-                conf_file.writelines([mon_hosts, "\n", keyring, "\n"])
+                # BUG#1865754 - ceph octopus requires "[global]" as part of
+                # the ini conf file that is generated.  The 'same_major=False'
+                # is based on the assumption that this is also the behaviour of
+                # ceph in future versions of ceph following octopus.
+                if versionutils.is_compatible(OCTOPUS_BASE_VERSION,
+                                              ceph_version,
+                                              same_major=False):
+                    config = ["[global]", "\n",
+                              mon_hosts, "\n", keyring, "\n"]
+                else:
+                    config = [mon_hosts, "\n", keyring, "\n"]
+
+                conf_file.writelines(config)
             return ceph_conf_path
         except IOError:
             msg = (_("Failed to write data to %s.") % (ceph_conf_path))
             raise exception.BrickException(msg=msg)
 
+    # output of 'ceph --version' on different versions of ceph
+    # ceph version 0.80.11 (8424145d49264624a3b0a204aedb127835161070)
+    # ceph version 10.2.11 (e4b061b47f07f583c92a050d9e84b1813a35671e)
+    # ceph version 12.2.13 (584a...) luminous (stable)
+    # ceph version 14.2.9 (581f...) nautilus (stable)
+    # ceph version 15.2.3 (d289...) octopus (stable)
+    # ceph version 14.2.11-99-gaf0268dc91 (af0268dc910f84b) nautilus (stable)
+    def _determine_ceph_client_version(self):
+        cmd = ['ceph', '--version']
+        (out, _) = self._execute(*cmd)
+        if '-' in out:
+            out = out[:out.index('-')]
+        return out.split(" ")[2]
+
     def _get_rbd_handle(self, connection_properties):
         try:
             user = connection_properties['auth_username']
@@ -110,6 +146,8 @@ class RBDConnector(base.BaseLinuxConnector):
             cluster_name = connection_properties['cluster_name']
             monitor_ips = connection_properties['hosts']
             monitor_ports = connection_properties['ports']
+            # NOTE: cinder no longer passes keyring data in the connection
+            # properties as of the victoria release.  See OSSN-0085.
             keyring = connection_properties.get('keyring')
         except (KeyError, ValueError):
             msg = _("Connect volume failed, malformed connection properties.")
@@ -130,7 +168,8 @@ class RBDConnector(base.BaseLinuxConnector):
 
         return rbd_handle
 
-    def _get_rbd_args(self, connection_properties):
+    @classmethod
+    def _get_rbd_args(cls, connection_properties, conf=None):
         try:
             user = connection_properties['auth_username']
             monitor_ips = connection_properties.get('hosts')
@@ -143,10 +182,14 @@ class RBDConnector(base.BaseLinuxConnector):
         if monitor_ips and monitor_ports:
             monitors = ["%s:%s" % (ip, port) for ip, port in
                         zip(
-                            self._sanitize_mon_hosts(monitor_ips),
+                            cls._sanitize_mon_hosts(monitor_ips),
                             monitor_ports)]
             for monitor in monitors:
                 args += ['--mon_host', monitor]
+
+        if conf:
+            args += ['--conf', conf]
+
         return args
 
     @staticmethod
@@ -160,33 +203,46 @@ class RBDConnector(base.BaseLinuxConnector):
         """
         return '/dev/rbd/{pool}/{volume}'.format(pool=pool, volume=volume)
 
-    @utils.trace
-    def connect_volume(self, connection_properties):
-        """Connect to a volume.
+    @classmethod
+    def create_non_openstack_config(cls, connection_properties):
+        """Get root owned Ceph's .conf file for non OpenStack usage."""
+        # If keyring info is missing then we are in OpenStack, nothing to do
+        keyring = connection_properties.get('keyring')
+        if not keyring:
+            return None
 
-        :param connection_properties: The dictionary that describes all
-                                      of the target volume attributes.
-        :type connection_properties: dict
-        :returns: dict
-        """
-        do_local_attach = connection_properties.get('do_local_attach',
-                                                    self.do_local_attach)
-
-        if do_local_attach:
-            # NOTE(e0ne): sanity check if ceph-common is installed.
-            cmd = ['which', 'rbd']
-            try:
-                self._execute(*cmd)
-            except putils.ProcessExecutionError:
-                msg = _("ceph-common package is not installed.")
-                LOG.error(msg)
-                raise exception.BrickException(message=msg)
-
-            # NOTE(e0ne): map volume to a block device
-            # via the rbd kernel module.
+        try:
+            user = connection_properties['auth_username']
             pool, volume = connection_properties['name'].split('/')
-            rbd_dev_path = RBDConnector.get_rbd_device_name(pool, volume)
+            cluster_name = connection_properties['cluster_name']
+            monitor_ips = connection_properties['hosts']
+            monitor_ports = connection_properties['ports']
+            keyring = connection_properties.get('keyring')
+        except (KeyError, ValueError):
+            msg = _("Connect volume failed, malformed connection properties.")
+            raise exception.BrickException(msg=msg)
 
+        conf = rbd_privsep.root_create_ceph_conf(monitor_ips, monitor_ports,
+                                                 str(cluster_name), user,
+                                                 keyring)
+        return conf
+
+    def _local_attach_volume(self, connection_properties):
+        # NOTE(e0ne): sanity check if ceph-common is installed.
+        try:
+            self._execute('which', 'rbd')
+        except putils.ProcessExecutionError:
+            msg = _("ceph-common package is not installed.")
+            LOG.error(msg)
+            raise exception.BrickException(message=msg)
+
+        # NOTE(e0ne): map volume to a block device
+        # via the rbd kernel module.
+        pool, volume = connection_properties['name'].split('/')
+        rbd_dev_path = self.get_rbd_device_name(pool, volume)
+        # If we are not running on OpenStack, create config file
+        conf = self.create_non_openstack_config(connection_properties)
+        try:
             if (
                 not os.path.islink(rbd_dev_path) or
                 not os.path.exists(os.path.realpath(rbd_dev_path))
@@ -195,7 +251,7 @@ class RBDConnector(base.BaseLinuxConnector):
                 # command introduced in ceph 13.0 (commit 6a57358add1157629a6d)
                 # when we drop support earlier versions
                 cmd = ['rbd', 'map', volume, '--pool', pool]
-                cmd += self._get_rbd_args(connection_properties)
+                cmd += self._get_rbd_args(connection_properties, conf)
                 self._execute(*cmd, root_helper=self._root_helper,
                               run_as_root=True)
             else:
@@ -215,13 +271,37 @@ class RBDConnector(base.BaseLinuxConnector):
                     'more information.',
                     {'vol': volume, 'dev': rbd_dev_path},
                 )
+        except Exception:
+            # Cleanup conf file on failure
+            with excutils.save_and_reraise_exception():
+                if conf:
+                    rbd_privsep.delete_if_exists(conf)
 
-            return {'path': rbd_dev_path, 'type': 'block'}
+        res = {'path': rbd_dev_path,
+               'type': 'block'}
+        if conf:
+            res['conf'] = conf
+        return res
+
+    @utils.trace
+    def connect_volume(self, connection_properties):
+        """Connect to a volume.
+
+        :param connection_properties: The dictionary that describes all
+                                      of the target volume attributes.
+        :type connection_properties: dict
+        :returns: dict
+        """
+        do_local_attach = connection_properties.get('do_local_attach',
+                                                    self.do_local_attach)
+
+        if do_local_attach:
+            return self._local_attach_volume(connection_properties)
 
         rbd_handle = self._get_rbd_handle(connection_properties)
         return {'path': rbd_handle}
 
-    def _find_root_device(self, connection_properties):
+    def _find_root_device(self, connection_properties, conf):
         """Find the underlying /dev/rbd* device for a mapping.
 
         Use the showmapped command to list all acive mappings and find the
@@ -237,7 +317,7 @@ class RBDConnector(base.BaseLinuxConnector):
         # command introduced in ceph 13.0 (commit 6a57358add1157629a6d)
         # when we drop support earlier versions
         cmd = ['rbd', 'showmapped', '--format=json']
-        cmd += self._get_rbd_args(connection_properties)
+        cmd += self._get_rbd_args(connection_properties, conf)
         (out, err) = self._execute(*cmd, root_helper=self._root_helper,
                                    run_as_root=True)
 
@@ -295,15 +375,18 @@ class RBDConnector(base.BaseLinuxConnector):
         do_local_attach = connection_properties.get('do_local_attach',
                                                     self.do_local_attach)
         if do_local_attach:
-            root_device = self._find_root_device(connection_properties)
+            conf = device_info.get('conf') if device_info else None
+            root_device = self._find_root_device(connection_properties, conf)
             if root_device:
                 # TODO(stephenfin): Update to the unified 'rbd device unmap'
                 # command introduced in ceph 13.0 (commit 6a57358add1157629a6d)
                 # when we drop support earlier versions
                 cmd = ['rbd', 'unmap', root_device]
-                cmd += self._get_rbd_args(connection_properties)
+                cmd += self._get_rbd_args(connection_properties, conf)
                 self._execute(*cmd, root_helper=self._root_helper,
                               run_as_root=True)
+                if conf:
+                    rbd_privsep.delete_if_exists(conf)
         else:
             if device_info:
                 rbd_handle = device_info.get('path', None)
@@ -311,13 +394,8 @@ class RBDConnector(base.BaseLinuxConnector):
                     fileutils.delete_if_exists(rbd_handle.rbd_conf)
                     rbd_handle.close()
 
-    def check_valid_device(self, path, run_as_root=True):
-        """Verify an existing RBD handle is connected and valid."""
-        rbd_handle = path
-
-        if rbd_handle is None:
-            return False
-
+    @staticmethod
+    def _check_valid_device(rbd_handle):
         original_offset = rbd_handle.tell()
 
         try:
@@ -331,6 +409,62 @@ class RBDConnector(base.BaseLinuxConnector):
 
         return True
 
+    def check_valid_device(self, path, run_as_root=True):
+        """Verify an existing RBD handle is connected and valid."""
+        if not path:
+            return False
+
+        # We can receive a file handle or a path to a device
+        if isinstance(path, str):
+            if run_as_root:
+                return rbd_privsep.check_valid_path(path)
+            else:
+                with open(path, 'rb') as rbd_handle:
+                    return self._check_valid_device(rbd_handle)
+
+        # For backward compatibility ignore run_as_root param with handles
+        return self._check_valid_device(path)
+
     def extend_volume(self, connection_properties):
-        # TODO(walter-boring): is this possible?
-        raise NotImplementedError
+        """Refresh local volume view and return current size in bytes."""
+        # Nothing to do, RBD attached volumes are automatically refreshed, but
+        # we need to return the new size for compatibility
+        do_local_attach = connection_properties.get('do_local_attach',
+                                                    self.do_local_attach)
+
+        if not do_local_attach:
+            handle = self._get_rbd_handle(connection_properties)
+            try:
+                # Handles should return absolute position on seek, but the RBD
+                # wrapper doesn't, so we need to call tell afterwards
+                handle.seek(0, 2)
+                return handle.tell()
+            finally:
+                fileutils.delete_if_exists(handle.rbd_conf)
+                handle.close()
+
+        # Create config file when we do the attach on the host and not the VM
+        conf = self.create_non_openstack_config(connection_properties)
+
+        try:
+            device_path = self._find_root_device(connection_properties, conf)
+        finally:
+            # If we have generated the config file we need to remove it
+            if conf:
+                try:
+                    rbd_privsep.delete_if_exists(conf)
+                except Exception as exc:
+                    LOG.warning(_('Could not remove config file %(filename)s: '
+                                  '%(exc)s'), {'filename': conf, 'exc': exc})
+
+        if not device_path:
+            msg = _('Cannot extend non mapped device.')
+            raise exception.BrickException(msg=msg)
+
+        device_name = os.path.basename(device_path)  # ie: rbd0
+        device_number = device_name[3:]  # ie: 0
+        # Get size from /sys/devices/rbd/0/size instead of
+        # /sys/class/block/rbd0/size because the latter isn't updated
+        with open('/sys/devices/rbd/' + device_number + '/size') as f:
+            size_bytes = f.read().strip()
+        return int(size_bytes)
